@@ -27,6 +27,8 @@ from sqlalchemy import (
     desc,
     update,
     or_,
+    case,
+    literal_column,
 )
 from sqlalchemy.exc import IntegrityError
 from sqlalchemy.ext.hybrid import hybrid_property
@@ -41,16 +43,22 @@ from sqlalchemy.engine.interfaces import ExecutionContext
 
 from cidc_schemas import prism, unprism, json_validation
 
-from .facets import get_facets_for_paths
+from .facets import get_facet_groups_for_paths, facet_groups_to_names
 from ..config.db import BaseModel
 from ..config.settings import (
     PAGINATION_PAGE_SIZE,
     MAX_PAGINATION_PAGE_SIZE,
     TESTING,
     INACTIVE_USER_DAYS,
+    GOOGLE_MAX_DOWNLOAD_PERMISSIONS,
 )
 from ..shared import emails
-from ..shared.gcloud_client import publish_artifact_upload, publish_upload_success
+from ..shared.gcloud_client import (
+    publish_artifact_upload,
+    publish_upload_success,
+    grant_download_access,
+    revoke_download_access,
+)
 
 
 def with_default_session(f):
@@ -237,6 +245,10 @@ class Users(CommonColumns):
         """Returns true if this user is a CIDC admin."""
         return self.role == CIDCRole.ADMIN.value
 
+    def is_nci_user(self) -> bool:
+        """Returns true if this user is an NCI Biobank user."""
+        return self.role == CIDCRole.NCI_BIOBANK_USER.value
+
     @with_default_session
     def update_accessed(self, session: Session, commit: bool = True):
         """Set this user's last system access to now."""
@@ -295,6 +307,10 @@ class Users(CommonColumns):
         return res
 
 
+class IAMException(Exception):
+    pass
+
+
 class Permissions(CommonColumns):
     __tablename__ = "permissions"
     __table_args__ = (
@@ -328,6 +344,66 @@ class Permissions(CommonColumns):
     trial_id = Column(String, nullable=False, index=True)
     upload_type = Column(String, nullable=False)
 
+    @with_default_session
+    def insert(self, session: Session, commit: bool = True, compute_etag: bool = True):
+        """
+        Insert this permission record into the database and add a corresponding IAM policy binding
+        on the GCS data bucket.
+
+        NOTE: values provided to the `commit` argument will be ignored. This method always commits.
+        """
+        grantee = Users.find_by_id(self.granted_to_user)
+        if grantee is None:
+            raise IntegrityError(
+                params=None,
+                statement=None,
+                orig=f"`granted_to_user` user must exist, but no user found with id {self.granted_to_user}",
+            )
+
+        # A user can only have 20 granular permissions at a time, due to GCS constraints
+        if (
+            len(Permissions.find_for_user(self.granted_to_user))
+            >= GOOGLE_MAX_DOWNLOAD_PERMISSIONS
+        ):
+            raise IntegrityError(
+                params=None,
+                statement=None,
+                orig=f"{grantee.email} has greater than or equal to the maximum number of allowed granular permissions ({GOOGLE_MAX_DOWNLOAD_PERMISSIONS}). Remove unused permissions to add others.",
+            )
+
+        # Always commit, because we don't want to grant IAM download unless this insert succeeds.
+        super().insert(session=session, commit=True, compute_etag=compute_etag)
+
+        try:
+            # Grant IAM permission in GCS only if db insert worked
+            grant_download_access(grantee.email, self.trial_id, self.upload_type)
+        except Exception as e:
+            # Delete the just-created permissions record
+            super().delete()
+            raise IAMException("IAM grant failed.") from e
+
+    @with_default_session
+    def delete(self, session: Session, commit: bool = True):
+        """
+        Delete this permission record from the database and revoke the corresponding IAM policy binding
+        on the GCS data bucket.
+
+        NOTE: values provided to the `commit` argument will be ignored. This method always commits.
+        """
+        grantee = Users.find_by_id(self.granted_to_user)
+        if grantee is None:
+            raise NoResultFound(f"no user with id {self.granted_to_user}")
+
+        try:
+            # Revoke IAM permission in GCS
+            revoke_download_access(grantee.email, self.trial_id, self.upload_type)
+        except Exception as e:
+            raise IAMException(
+                "IAM revoke failed, and permission db record not removed."
+            ) from e
+
+        super().delete(session=session, commit=True)
+
     @staticmethod
     @with_default_session
     def find_for_user(user_id: int, session: Session) -> List:
@@ -346,6 +422,29 @@ class Permissions(CommonColumns):
             .first()
         )
 
+    @staticmethod
+    @with_default_session
+    def grant_all_iam_permissions(session: Session):
+        Permissions._change_all_iam_permissions(grant=True, session=session)
+
+    @staticmethod
+    @with_default_session
+    def revoke_all_iam_permissions(session: Session):
+        Permissions._change_all_iam_permissions(grant=False, session=session)
+
+    @staticmethod
+    @with_default_session
+    def _change_all_iam_permissions(grant: bool, session: Session):
+        perms = Permissions.list(page_size=Permissions.count(), session=session)
+        for perm in perms:
+            user = Users.find_by_id(perm.granted_to_user)
+            if user.is_admin() or user.is_nci_user():
+                continue
+            if grant:
+                grant_download_access(user.email, perm.trial_id, perm.upload_type)
+            else:
+                revoke_download_access(user.email, perm.trial_id, perm.upload_type)
+
 
 class ValidationMultiError(Exception):
     """Holds multiple jsonschema.ValidationErrors"""
@@ -772,6 +871,18 @@ class DownloadableFiles(CommonColumns):
     def file_ext(cls):
         return func.substring(cls.object_url, cls.FILE_EXT_REGEX)
 
+    @hybrid_property
+    def data_category(self):
+        return facet_groups_to_names.get(self.facet_group)
+
+    @data_category.expression
+    def data_category(cls):
+        return DATA_CATEGORY_CASE_CLAUSE
+
+    @property
+    def flat_object_url(self):
+        return self.object_url.replace("/", "_")
+
     @staticmethod
     def build_file_filter(
         trial_ids: List[str] = [], facets: List[List[str]] = [], user: Users = None
@@ -794,11 +905,10 @@ class DownloadableFiles(CommonColumns):
         if trial_ids:
             file_filters.append(DownloadableFiles.trial_id.in_(trial_ids))
         if facets:
-            facet_filters = get_facets_for_paths(
-                DownloadableFiles.object_url.like, facets
-            )
-            file_filters.append(or_(*facet_filters))
-        if user and not user.is_admin():
+            facet_groups = get_facet_groups_for_paths(facets)
+            file_filters.append(DownloadableFiles.facet_group.in_(facet_groups))
+        # Admins and NCI biobank users can view all files
+        if user and not user.is_admin() and not user.is_nci_user():
             permissions = Permissions.find_for_user(user.id)
             perm_set = [(p.trial_id, p.upload_type) for p in permissions]
             file_tuples = tuple_(
@@ -914,3 +1024,10 @@ class DownloadableFiles(CommonColumns):
         the given GCS object url.
         """
         return session.query(DownloadableFiles).filter_by(object_url=object_url).one()
+
+
+# Query clause for computing a downloadable file's data category.
+# Used above in the DownloadableFiles.data_category computed property.
+DATA_CATEGORY_CASE_CLAUSE = case(
+    [(DownloadableFiles.facet_group == k, v) for k, v in facet_groups_to_names.items()]
+)
